import abc
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from flax.training.common_utils import shard_prng_key
from jax import (
    jit,
    lax,
    pmap,
    random,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
import jax.scipy as jsp
from jax.tree_util import tree_map

from fortuna.data.loader import (
    DataLoader,
    DeviceDimensionAugmentedLoader,
    InputsLoader,
    TargetsLoader,
)
from fortuna.prob_model.posterior.base import Posterior
from fortuna.typing import (
    Array,
    Batch,
    CalibMutable,
    CalibParams,
)
from fortuna.utils.random import WithRNG


class Predictive(WithRNG):
    def __init__(self, posterior: Posterior):
        """
        Predictive distribution abstract class.

        Parameters
        ----------
        posterior : Posterior
             A posterior distribution object.
        """
        self.likelihood = posterior.joint.likelihood
        self.posterior = posterior

    ####
    def ensemble_log_prob(
        self,
        data_loader: DataLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Compute the log-likelihood at each posterior sample, that is

        .. math::
            \log p(y|x, theta^{(i)}),

        where:
         - :math:`x` is an observed input variable;
         - :math:`y` is an observed target variable;
         - :math:`theta^{(i)}` is a sample from the posterior.

        Parameters
        ----------
        data_loader : DataLoader
            A data loader.
        n_posterior_samples : int
            Number of posterior samples to draw in order to compute the log -ikelihood.
            that would be produced using the posterior distribution state.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An array of log-likelihood values at each posterior sample for each data point.
        """

        if rng is None:
            rng = self.rng.get()

        log_probs = self._loop_fun_through_data_loader(
            self._batched_ensemble_log_prob,
            data_loader,
            n_posterior_samples,
            rng,
            distribute,
            **kwargs,
        )

        return log_probs.swapaxes(0, 1)

    def _batched_ensemble_log_prob(
        self,
        batch: Batch,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, dict]]:
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_posterior_samples)

        def _lik_log_batched_prob(key):
            sample = self.posterior.sample(inputs=batch[0], rng=key)
            return self.likelihood._batched_log_prob(
                sample.params,
                batch,
                mutable=sample.mutable,
                calib_params=sample.calib_params,
                calib_mutable=sample.calib_mutable,
                **kwargs,
            )

        return lax.map(_lik_log_batched_prob, keys).swapaxes(
            0, 1
        )  # notice that we use lax.map and not vmap to avoid parallel posterior sampling

    ####
    def log_prob(
        self,
        data_loader: DataLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive log-probability density function (a.k.a. log-pdf), that is

        .. math::
            \log p(y|x, \mathcal{D}),

        where:
         - :math:`x` is an observed input variable;
         - :math:`y` is an observed target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Parameters
        ----------
        data_loader : DataLoader
            A data loader.
        n_posterior_samples : int
            Number of posterior samples to draw in order to approximate the predictive log-pdf.
            that would be produced using the posterior distribution state.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive log-pdf for each data point.
        """
        if rng is None:
            rng = self.rng.get()

        return self._loop_fun_through_data_loader(
            self._batched_log_prob,
            data_loader,
            n_posterior_samples,
            rng,
            distribute,
            **kwargs,
        )

    def _batched_log_prob(
        self,
        batch: Batch,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, dict]]:
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_posterior_samples)

        def _lik_log_batched_prob(key):
            sample = self.posterior.sample(inputs=batch[0], rng=key)
            return self.likelihood._batched_log_prob(
                sample.params,
                batch,
                mutable=sample.mutable,
                calib_params=sample.calib_params,
                calib_mutable=sample.calib_mutable,
                **kwargs,
            )

        return jsp.special.logsumexp(
            lax.map(_lik_log_batched_prob, keys), axis=0
        ) - jnp.log(n_posterior_samples)

    def _batched_log_joint_prob(
        self,
        batch: Batch,
        n_data: int,
        n_posterior_samples: int = 30,
        return_aux: Optional[List[str]] = None,
        ensemble_outputs: Optional[jnp.ndarray] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]:
        if return_aux is None:
            return_aux = []
        supported_aux = ["outputs", "calib_mutable"]
        unsupported_aux = [s for s in return_aux if s not in supported_aux]
        if sum(unsupported_aux) > 0:
            raise Exception(
                """The auxiliary objects {} are unknown. Please make sure that all elements of `return_aux`
                            belong to the following list: {}""".format(
                    unsupported_aux, supported_aux
                )
            )

        if ensemble_outputs is None:
            if rng is None:
                rng = self.rng.get()
            keys = random.split(rng, n_posterior_samples)

            def _lik_log_joint_prob(key):
                sample = self.posterior.sample(inputs=batch[0], rng=key)
                return self.likelihood._batched_log_joint_prob(
                    sample.params,
                    batch,
                    n_data,
                    mutable=sample.mutable,
                    return_aux=return_aux,
                    calib_params=calib_params
                    if calib_params is not None
                    else sample.calib_params,
                    calib_mutable=calib_mutable
                    if calib_mutable is not None
                    else sample.calib_mutable,
                    **kwargs,
                )

            outs = lax.map(_lik_log_joint_prob, keys)
        else:
            if calib_params is None:
                d = self.posterior.state.extract_calib_keys()
                calib_params = d["calib_params"]
                calib_mutable = d["calib_mutable"]

            def _lik_log_joint_prob(outputs):
                return self.likelihood._batched_log_joint_prob(
                    None,
                    batch,
                    n_data,
                    mutable=None,
                    outputs=outputs,
                    return_aux=return_aux,
                    calib_params=calib_params,
                    calib_mutable=calib_mutable,
                    **kwargs,
                )

            outs = lax.map(_lik_log_joint_prob, ensemble_outputs)

        if len(return_aux) > 0:
            ensemble_log_lik, aux = outs
            aux = tree_map(lambda v: jnp.mean(v, 0), aux)
        else:
            ensemble_log_lik = outs
        log_pred = jsp.special.logsumexp(ensemble_log_lik) - jnp.log(
            len(ensemble_log_lik)
        )
        if len(return_aux) > 0:
            return log_pred, aux
        return log_pred

    def _batched_negative_log_joint_prob(
        self,
        batch: Batch,
        n_data: int,
        n_posterior_samples: int = 30,
        return_aux: Optional[List[str]] = None,
        ensemble_outputs: Optional[jnp.ndarray] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]:
        outs = self._batched_log_joint_prob(
            batch,
            n_data,
            n_posterior_samples,
            return_aux,
            ensemble_outputs,
            calib_params,
            calib_mutable,
            rng,
            **kwargs,
        )
        if len(return_aux) > 0:
            loss, aux = outs
            loss *= -1
            return loss, aux
        return -outs

    def sample(
        self,
        inputs_loader: InputsLoader,
        n_target_samples: int = 1,
        return_aux: Optional[List[str]] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        **kwargs,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]]:
        r"""
        Sample from an approximation of the predictive distribution for each input data point, that is

        .. math::
            y^{(i)}\sim p(\cdot|x, \mathcal{D}),

        where:
         - :math:`x` is an observed input variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`y^{(i)}` is a sample of the target variable for the input :math:`x`.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_target_samples : int
            Number of target samples to sample for each input data point.
        return_aux : Optional[List[str]]
            Return auxiliary objects. We currently support 'outputs'.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        Union[jnp.ndarray, Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]]
            Samples for each input data point. Optionally, an auxiliary object is returned.
        """
        if not rng:
            rng = self.rng.get()

        def fun(_inputs):
            return self._batched_sample(
                _inputs, n_target_samples, return_aux, rng, **kwargs
            )

        if distribute:
            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)
            fun = pmap(fun)
            if return_aux is None or len(return_aux) == 0:
                return jnp.concatenate(
                    [
                        self.likelihood._unshard_array(fun(inputs))
                        for inputs in inputs_loader
                    ],
                    1,
                )
            else:
                samples, aux_outputs = [], []
                for inputs in inputs_loader:
                    _samples, _aux = fun(inputs)
                    samples.append(self.likelihood._unshard_array(_samples))
                    if "outputs" in _aux:
                        aux_outputs.append(
                            self.likelihood._unshard_array(_aux["outputs"])
                        )
                samples = jnp.concatenate(samples, axis=0)
                aux = dict()
                if "outputs" in aux:
                    aux["outputs"] = jnp.concatenate(aux_outputs, axis=0)
                return samples, aux
        else:
            fun = jit(fun)
            if return_aux is None or len(return_aux) == 0:
                return jnp.concatenate([fun(inputs) for inputs in inputs_loader], 1)
            else:
                samples, aux_outputs = [], []
                for inputs in inputs_loader:
                    _samples, _aux = fun(inputs)
                    samples.append(_samples)
                    if "outputs" in _aux:
                        aux_outputs.append(_aux["outputs"])
                samples = jnp.concatenate(samples, axis=0)
                aux = dict()
                if "outputs" in aux:
                    aux["outputs"] = jnp.concatenate(aux_outputs, axis=0)
                return samples, aux

    def _batched_sample(
        self,
        inputs: Array,
        n_target_samples: int = 1,
        return_aux: Optional[List[str]] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs,
    ) -> jnp.ndarray:
        if return_aux is None:
            return_aux = []

        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_target_samples)

        def _sample(key):
            key1, key2 = random.split(key, 2)
            _post_sample = self.posterior.sample(inputs=inputs, rng=key1)
            outs = self.likelihood._batched_sample(
                1,
                _post_sample.params,
                inputs,
                mutable=_post_sample.mutable,
                calib_params=_post_sample.calib_params,
                calib_mutable=_post_sample.calib_mutable,
                return_aux=return_aux,
                rng=key2,
                **kwargs,
            )
            if len(return_aux) > 0:
                _samples, aux = outs
                return _samples.squeeze(0), aux
            return outs.squeeze(0)

        return lax.map(_sample, keys)

    def sample_calibrated_outputs(
        self,
        inputs_loader: InputsLoader,
        n_output_samples: int = 1,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Sample parameters from the posterior distribution state and compute calibrated outputs.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_output_samples : int
            Number of output samples to draw for each input.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            Samples of calibrated outputs.
        """
        if rng is None:
            rng = self.rng.get()

        return self._loop_ensemble_fun_through_inputs_loader(
            self._sample_batched_calibrated_outputs,
            inputs_loader,
            n_output_samples,
            rng,
            distribute,
        )

    def _sample_batched_calibrated_outputs(
        self,
        inputs: Array,
        n_output_samples: int = 1,
        rng: Optional[PRNGKeyArray] = None,
    ) -> jnp.ndarray:
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_output_samples)

        def _sample(key):
            sample = self.posterior.sample(inputs=inputs, rng=key)
            return self.likelihood._get_batched_calibrated_outputs(
                params=sample.params,
                inputs=inputs,
                mutable=sample.mutable,
                calib_params=sample.calib_params,
                calib_mutable=sample.calib_mutable,
            )

        return lax.map(_sample, keys)

    def _sample_outputs(
        self,
        inputs_loader: InputsLoader,
        n_output_samples: int = 1,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        if rng is None:
            rng = self.rng.get()

        return self._loop_fun_through_inputs_loader(
            self._sample_batched_outputs,
            inputs_loader,
            n_output_samples,
            rng,
            distribute,
        )

    def _sample_batched_outputs(
        self,
        inputs: Array,
        n_output_samples: int = 1,
        rng: Optional[PRNGKeyArray] = None,
    ) -> jnp.ndarray:
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_output_samples)

        def _sample(key):
            sample = self.posterior.sample(inputs=inputs, rng=key)
            return self.likelihood.model_manager.apply(
                params=sample.params, inputs=inputs, mutable=sample.mutable
            )

        return lax.map(_sample, keys)

    def _sample_outputs_loader(
        self,
        inputs_loader: InputsLoader,
        n_output_samples: int = 1,
        rng: Optional[PRNGKeyArray] = None,
        return_size: bool = False,
        distribute: bool = True,
    ) -> Union[TargetsLoader, Tuple[TargetsLoader, int]]:
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_output_samples)

        if distribute:
            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)

        def _sample(key, _inputs):
            sample = self.posterior.sample(inputs=_inputs, rng=key)
            return self.likelihood.model_manager.apply(
                params=sample.params, inputs=_inputs, mutable=sample.mutable
            )

        _sample = pmap(_sample) if distribute else jit(_sample)

        iterable = []
        size = 0
        for inputs in inputs_loader:
            size += (
                inputs.shape[0]
                if not isinstance(inputs, dict)
                else inputs[list(inputs.keys())[0]].shape[0]
            )
            if distribute:
                outputs = jnp.stack(
                    list(map(lambda key: _sample(shard_prng_key(key), inputs), keys))
                )
                outputs = self._unshard_ensemble_arrays(outputs)
            else:
                outputs = lax.map(lambda key: _sample(key, inputs), keys)
            iterable.append(outputs)
        iterable = TargetsLoader.from_iterable(iterable=iterable)
        if return_size:
            return iterable, size
        return iterable

    def mean(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive mean of the target variable, that is

        .. math::
            \mathbb{E}_{Y|x, \mathcal{D}}[Y],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive mean for each input.
        """
        if rng is None:
            rng = self.rng.get()

        return self._loop_fun_through_inputs_loader(
            self._batched_mean, inputs_loader, n_posterior_samples, rng, distribute
        )

    def _batched_mean(
        self,
        inputs: Array,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
    ) -> jnp.ndarray:
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_posterior_samples)

        def fun(i, _curr_sum):
            _sample = self.posterior.sample(inputs=inputs, rng=keys[i])
            _curr_sum += self.likelihood._batched_mean(
                _sample.params,
                inputs,
                _sample.mutable,
                calib_params=_sample.calib_params,
                calib_mutable=_sample.calib_mutable,
            )
            return _curr_sum

        curr_sum = fun(0, 0.0)
        curr_sum = lax.fori_loop(1, n_posterior_samples, fun, curr_sum)
        return curr_sum / n_posterior_samples

    @abc.abstractmethod
    def mode(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        means: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive mode of the target variable, that is

        .. math::
            \text{argmax}_y\ p(y|x, \mathcal{D}),

        where:
         - :math:`x` is an observed input variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`y` is the target variable to optimize upon.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        means : Optional[jnp.ndarray] = None
            An estimate of the predictive mean.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive mode for each input.
        """
        pass

    def aleatoric_variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive aleatoric variance of the target variable, that is

        .. math::
            \text{Var}_{W|\mathcal{D}}[\mathbb{E}_{Y|W, x}[Y]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive aleatoric variance for each input.
        """
        if rng is None:
            rng = self.rng.get()

        return self._loop_fun_through_inputs_loader(
            self._batched_aleatoric_variance,
            inputs_loader,
            n_posterior_samples,
            rng,
            distribute,
        )

    def _batched_aleatoric_variance(
        self,
        inputs: Array,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
    ) -> jnp.ndarray:
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_posterior_samples)

        def fun(i, _curr_sum):
            _sample = self.posterior.sample(inputs=inputs, rng=keys[i])
            _curr_sum += self.likelihood._batched_variance(
                _sample.params,
                inputs,
                _sample.mutable,
                calib_params=_sample.calib_params,
                calib_mutable=_sample.calib_mutable,
            )
            return _curr_sum

        curr_sum = fun(0, 0.0)
        curr_sum = lax.fori_loop(1, n_posterior_samples, fun, curr_sum)
        return curr_sum / n_posterior_samples

    def epistemic_variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive epistemic variance of the one-hot encoded target variable, that is

        .. math::
            \mathbb{E}_{W|D}[\text{Var}_{Y|W, x}[Y]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive epistemic variance for each input.
        """
        if rng is None:
            rng = self.rng.get()

        return self._loop_fun_through_inputs_loader(
            self._batched_epistemic_variance,
            inputs_loader,
            n_posterior_samples,
            rng,
            distribute,
        )

    def _batched_epistemic_variance(
        self,
        inputs: Array,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
    ) -> jnp.ndarray:
        if rng is None:
            rng = self.rng.get()
        keys = random.split(rng, n_posterior_samples)

        def fun(i, variables):
            _curr_sum, _curr_sum_sq = variables
            _sample = self.posterior.sample(inputs=inputs, rng=keys[i])
            mean = self.likelihood._batched_mean(
                _sample.params,
                inputs,
                _sample.mutable,
                calib_params=_sample.calib_params,
                calib_mutable=_sample.calib_mutable,
            )
            _curr_sum += mean
            _curr_sum_sq += mean**2
            return _curr_sum, _curr_sum_sq

        curr_sum, curr_sum_sq = fun(0, (0.0, 0.0))
        curr_sum, curr_sum_sq = lax.fori_loop(
            1, n_posterior_samples, fun, (curr_sum, curr_sum_sq)
        )
        return jnp.maximum(
            0, curr_sum_sq / n_posterior_samples - (curr_sum / n_posterior_samples) ** 2
        )

    def variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        aleatoric_variances: Optional[jnp.ndarray] = None,
        epistemic_variances: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive variance of the target variable, that is

        .. math::
            \text{Var}_{Y|x, D}[Y],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Note that the predictive variance above corresponds to the sum of its aleatoric and epistemic components.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        aleatoric_variances: Optional[jnp.ndarray]
            An estimate of the aleatoric predictive variance for each input.
        epistemic_variances: Optional[jnp.ndarray]
            An estimate of the epistemic predictive variance for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive variance for each input.
        """
        if rng is None:
            rng = self.rng.get()
        if aleatoric_variances is None:
            rng, key = random.split(rng)
            aleatoric_variances = self.aleatoric_variance(
                inputs_loader=inputs_loader,
                n_posterior_samples=n_posterior_samples,
                rng=key,
                distribute=distribute,
            )
        if epistemic_variances is None:
            rng, key = random.split(rng)
            epistemic_variances = self.epistemic_variance(
                inputs_loader=inputs_loader,
                n_posterior_samples=n_posterior_samples,
                rng=key,
                distribute=distribute,
            )
        return aleatoric_variances + epistemic_variances

    def std(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        variances: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive standard deviation of the target variable, that is

        .. math::
            \text{Var}_{Y|x, D}[Y],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        variances: Optional[jnp.ndarray]
            An estimate of the predictive variance.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive standard deviation for each input.
        """
        if variances is None:
            variances = self.variance(
                inputs_loader=inputs_loader,
                n_posterior_samples=n_posterior_samples,
                rng=rng,
                distribute=distribute,
            )
        return jnp.sqrt(variances)

    @staticmethod
    def _unshard_ensemble_arrays(arr: Array) -> Array:
        arr = arr.swapaxes(1, 2)
        arr = arr.reshape((arr.shape[0] * arr.shape[1],) + arr.shape[2:])
        return arr.swapaxes(0, 1)

    def _loop_fun_through_inputs_loader(
        self,
        fun: Callable,
        inputs_loader: InputsLoader,
        n_posterior_samples: int,
        rng: PRNGKeyArray,
        distribute: bool = True,
        **kwargs,
    ) -> Array:
        def fun2(_inputs):
            return fun(_inputs, n_posterior_samples, rng, **kwargs)

        if distribute:
            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)
            fun2 = pmap(fun2)
            return jnp.concatenate(
                [
                    self.likelihood._unshard_array(fun2(inputs))
                    for inputs in inputs_loader
                ],
                0,
            )
        fun2 = jit(fun2)
        return jnp.concatenate([fun2(inputs) for inputs in inputs_loader], 0)

    def _loop_fun_through_data_loader(
        self,
        fun: Callable,
        data_loader: DataLoader,
        n_posterior_samples: int,
        rng: PRNGKeyArray,
        distribute: bool = True,
        **kwargs,
    ) -> Array:
        def fun2(_batch):
            return fun(_batch, n_posterior_samples, rng, **kwargs)

        if distribute:
            data_loader = DeviceDimensionAugmentedLoader(data_loader)
            fun2 = pmap(fun2)
            return jnp.concatenate(
                [self.likelihood._unshard_array(fun2(batch)) for batch in data_loader],
                0,
            )
        fun2 = jit(fun2)
        return jnp.concatenate([fun2(batch) for batch in data_loader], 0)

    def _loop_ensemble_fun_through_inputs_loader(
        self,
        fun: Callable,
        inputs_loader: InputsLoader,
        n_posterior_samples: int,
        rng: PRNGKeyArray,
        distribute: bool = True,
        **kwargs,
    ) -> Array:
        def fun2(_inputs):
            return fun(_inputs, n_posterior_samples, rng, **kwargs)

        if distribute:
            inputs_loader = DeviceDimensionAugmentedLoader(inputs_loader)
            fun2 = pmap(fun2)
            return jnp.concatenate(
                [
                    self._unshard_ensemble_arrays(fun2(inputs))
                    for inputs in inputs_loader
                ],
                1,
            )
        fun2 = jit(fun2)
        return jnp.concatenate([fun2(inputs) for inputs in inputs_loader], 1)



from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.deep_ensemble import DEEP_ENSEMBLE_NAME


class DeepEnsemblePosteriorApproximator(PosteriorApproximator):
    def __init__(self, ensemble_size: int = 5):
        """
        Deep ensemble posterior approximator. It is responsible to define how the posterior distribution is
        approximated.

        Parameters
        ----------
        ensemble_size : int
            The size of the ensemble.
        """
        self.ensemble_size = ensemble_size

    def __str__(self):
        return DEEP_ENSEMBLE_NAME



from typing import (
    List,
    Optional,
)

from fortuna.calib_model.config.callback import Callback
from fortuna.calib_model.config.checkpointer import Checkpointer
from fortuna.calib_model.config.hyperparameters import Hyperparameters
from fortuna.calib_model.config.monitor import Monitor
from fortuna.calib_model.config.optimizer import Optimizer
from fortuna.calib_model.config.processor import Processor


class Config:
    def __init__(
        self,
        optimizer: Optimizer = Optimizer(),
        checkpointer: Checkpointer = Checkpointer(),
        monitor: Monitor = Monitor(),
        processor: Processor = Processor(),
        hyperparameters: Hyperparameters = Hyperparameters(),
        callbacks: Optional[List[Callback]] = None,
    ):
        """
        Configure the posterior distribution fitting.

        Parameters
        ----------
        optimizer: Optimizer
            It defines the optimization specifics.
        checkpointer: Checkpointer
            It handles saving and restoring checkpoints.
        monitor: Monitor
            It monitors training progress and might induce early stopping.
        processor: Processor
            It processes where computation takes place.
        hyperparameters: Hyperparameters
            It defines other hyperparameters that may be needed during model's training.
        callbacks:  Optional[List[FitCallback]]
            A list of user-defined callbacks to be called during training.
            Callbacks run sequentially in the order defined by the user.
        """
        self.optimizer = optimizer
        self.checkpointer = checkpointer
        self.monitor = monitor
        self.processor = processor
        self.hyperparameters = hyperparameters
        self.callbacks = callbacks



from typing import (
    Dict,
    Optional,
)

import flax.linen as nn
import numpy as np

from fortuna.data.loader import DataLoader
from fortuna.likelihood.regression import RegressionLikelihood
from fortuna.model.model_manager.regression import RegressionModelManager
from fortuna.model_editor.base import ModelEditor
from fortuna.output_calibrator.output_calib_manager.base import OutputCalibManager
from fortuna.output_calibrator.regression import RegressionTemperatureScaler
from fortuna.prob_model.base import ProbModel
from fortuna.prob_model.calib_config.base import CalibConfig
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.posterior_approximations import (
    PosteriorApproximations,
)
from fortuna.prob_model.posterior.swag.swag_approximator import (
    SWAGPosteriorApproximator,
)
from fortuna.prob_model.predictive.regression import RegressionPredictive
from fortuna.prob_model.prior import IsotropicGaussianPrior
from fortuna.prob_model.prior.base import Prior
from fortuna.prob_output_layer.regression import RegressionProbOutputLayer
from fortuna.typing import Status


class ProbRegressor(ProbModel):
    def __init__(
        self,
        model: nn.Module,
        likelihood_log_variance_model: nn.Module,
        prior: Prior = IsotropicGaussianPrior(),
        posterior_approximator: PosteriorApproximator = SWAGPosteriorApproximator(),
        output_calibrator: Optional[nn.Module] = RegressionTemperatureScaler(),
        model_editor: Optional[ModelEditor] = None,
        seed: int = 0,
    ):
        r"""
        A probabilistic regressor class.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. It characterizes the mean model
            of the likelihood function. The outputs must belong to the same space as the target variables.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`\mu(w, x)`.
        likelihood_log_variance_model: nn.Module
            A model characterizing the log-variance of a Gaussian likelihood function. The outputs must belong to the
            same space as the target variables. Let :math:`x` be input variables and :math:`w` the random model
            parameters. Then the model is described by a function :math:`\log\sigma^2(w, x)`.
        prior : Prior
            A prior distribution object. The default is an isotropic standard Gaussian. Let :math:`w` be the random
            model parameters. Then the prior is defined by a distribution :math:`p(w)`.
        posterior_approximator : PosteriorApproximator
            A posterior approximation method. The default method is SWAG.
        output_calibrator : Optional[nn.Module]
            An output calibrator object. The default is temperature scaling for regression, which inflates the variance
            of the likelihood with a scalar temperature parameter. Given outputs :math:`o` of the model manager, the
            output calibrator is described by a function :math:`g(\phi, o)`, where `phi` are deterministic
            calibration parameters.
        model_editor : ModelEditor
            A model_editor objects. It takes the forward pass and transforms the outputs.
        seed: int
            A random seed.

        Attributes
        ----------
        model : nn.Module
            See `model` in `Parameters`.
        model_manager : RegressionModelManager
            This object orchestrates the model's forward pass. Given a mean model :math:`\mu(w, x)` and a log-variance
            model :math:`\log\sigma^2`, the model manager concatenates the two into
            :math:`f(w, x)=[\mu(w, x), \log\sigma^2(w, x)]`.
        output_calibrator : nn.Module
            See `output_calibrator` in `Parameters`.
        prob_output_layer : RegressionProbOutputLayer
            This object characterizes the distribution of the target variable given the calibrated outputs. It is
            defined by :math:`p(y|\omega)=\text{Categorical}(p=softmax(\omega))`, where :math:`\omega` denote the
            calibrated outputs and :math:`y` denotes a target variable.
        likelihood : RegressionLikelihood
            The likelihood function. This is defined by
            :math:`p(y|w, \phi, x) = \text{Categorical}(p=\text{softmax}(g(\phi, f(w, x)))`.
        prior : Prior
            See `prior` in `Parameters`.
        joint : Joint
            This object describes the joint distribution of the target variables and the random parameters
            given the input variables and the calibration parameters, that is :math:`p(y, w|x, \phi)`.
        posterior_approximator : PosteriorApproximator
            See `posterior_approximator` in `Parameters`.
        posterior : Posterior
            This is the posterior approximation of the random parameters given the training data and the
            calibration parameters, that is :math:`p(w|\mathcal{D}, \phi)`, where :math:`\mathcal{D}` denotes the
            training data set and :math:`\phi` the calibration parameters.
        predictive : RegressionPredictive
            This denotes the predictive distribution, that is :math:`p(y|\phi, x, \mathcal{D})`. Its statistics are
            approximated via a Monte Carlo approach by sampling from the posterior approximation.
        """
        self.model = model
        self.lik_log_var = likelihood_log_variance_model
        self.prior = prior
        self.output_calibrator = output_calibrator

        self.model_manager = RegressionModelManager(
            model, likelihood_log_variance_model, model_editor=model_editor
        )
        self.output_calib_manager = OutputCalibManager(
            output_calibrator=output_calibrator
        )
        self.prob_output_layer = RegressionProbOutputLayer()

        self.likelihood = RegressionLikelihood(
            self.model_manager, self.prob_output_layer, self.output_calib_manager
        )
        self.joint = Joint(self.prior, self.likelihood)

        self.posterior = getattr(
            PosteriorApproximations, posterior_approximator.__str__()
        ).value(joint=self.joint, posterior_approximator=posterior_approximator)
        self.predictive = RegressionPredictive(self.posterior)

        super().__init__(seed=seed)

    def _check_output_dim(self, data_loader: DataLoader):
        output_dim = 0
        for x, y in data_loader:
            input_shape = x.shape[1:]
            output_dim = y.shape[1]
            break
        if output_dim == 0:
            raise ValueError(
                """`_data_loader` is either empty or incorrectly constructed."""
            )
        s = self.joint.init(input_shape)
        outputs = self.model_manager.apply(
            params=s.params, inputs=np.zeros((1,) + input_shape), mutable=s.mutable
        )
        if outputs.shape[1] != 2 * output_dim:
            raise ValueError(
                f"""The outputs dimension of both `model` and `likelihood_log_variance_model` must be the same as
                the dimension of the target variables in `_data_loader`. However, {outputs.shape[1] // 2} and
                {output_dim} were found, respectively."""
            )

    def train(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        calib_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        calib_config: CalibConfig = CalibConfig(),
        **fit_kwargs,
    ) -> Dict[str, Status]:
        self._check_output_dim(train_data_loader)
        return super().train(
            train_data_loader,
            val_data_loader,
            calib_data_loader,
            fit_config,
            calib_config,
            **fit_kwargs,
        )

    def calibrate(
        self,
        calib_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        calib_config: CalibConfig = CalibConfig(),
    ) -> Status:
        """
        Calibrate the probabilistic classifier.

        Parameters
        ----------
        calib_data_loader : DataLoader
            A calibration data loader.
        val_data_loader : DataLoader
            A validation data loader.
        calib_config : CalibConfig
            An object to configure the calibration.

        Returns
        -------
        Status
            A calibration status object. It provides information about the calibration.
        """
        self._check_output_dim(calib_data_loader)
        if val_data_loader is not None:
            self._check_output_dim(val_data_loader)
        return super()._calibrate(
            uncertainty_fn=calib_config.monitor.uncertainty_fn
            if calib_config.monitor.uncertainty_fn is not None
            else self.prob_output_layer.variance,
            calib_data_loader=calib_data_loader,
            val_data_loader=val_data_loader,
            calib_config=calib_config,
        )



from typing import (
    List,
    Optional,
)

from fortuna.prob_model.fit_config.callback import FitCallback
from fortuna.prob_model.fit_config.checkpointer import FitCheckpointer
from fortuna.prob_model.fit_config.hyperparameters import FitHyperparameters
from fortuna.prob_model.fit_config.monitor import FitMonitor
from fortuna.prob_model.fit_config.optimizer import FitOptimizer
from fortuna.prob_model.fit_config.processor import FitProcessor


class FitConfig:
    def __init__(
        self,
        optimizer: FitOptimizer = FitOptimizer(),
        checkpointer: FitCheckpointer = FitCheckpointer(),
        monitor: FitMonitor = FitMonitor(),
        processor: FitProcessor = FitProcessor(),
        hyperparameters: FitHyperparameters = FitHyperparameters(),
        callbacks: Optional[List[FitCallback]] = None,
    ):
        """
        Configure the posterior distribution fitting.

        Parameters
        ----------
        optimizer: FitOptimizer
            It defines the optimization specifics.
        checkpointer: FitCheckpointer
            It handles saving and restoring checkpoints.
        monitor: FitMonitor
            It monitors training progress and might induce early stopping.
        processor: FitProcessor
            It processes where computation takes place.
        hyperparameters: FitHyperparameters
            It defines other hyperparameters that may be needed during model's training.
        callbacks:  Optional[List[FitCallback]]
            A list of user-defined callbacks to be called during training.
            Callbacks run sequentially in the order defined by the user.
        """
        self.optimizer = optimizer
        self.checkpointer = checkpointer
        self.monitor = monitor
        self.processor = processor
        self.hyperparameters = hyperparameters
        self.callbacks = callbacks



from typing import Optional

import jax.numpy as jnp

from fortuna.output_calib_model.predictive.base import Predictive
from fortuna.output_calibrator.output_calib_manager.base import OutputCalibManager
from fortuna.prob_output_layer.classification import ClassificationProbOutputLayer
from fortuna.typing import Array


class ClassificationPredictive(Predictive):
    def __init__(
        self,
        output_calib_manager: OutputCalibManager,
        prob_output_layer: ClassificationProbOutputLayer,
    ):
        super().__init__(
            output_calib_manager=output_calib_manager,
            prob_output_layer=prob_output_layer,
        )

    def mean(self, outputs: Array, calibrated: bool = True, **kwargs) -> jnp.ndarray:
        """
        Estimate the mean of the one-hot encoded target variable given the output, with respect to the predictive
        distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated mean for each output.
        """
        return super().mean(outputs, calibrated, **kwargs)

    def mode(self, outputs: Array, calibrated: bool = True, **kwargs) -> jnp.ndarray:
        """
        Estimate the mode of the one-hot encoded target variable given the output, with respect to the predictive
        distribution.

        Parameters
        ----------
        outputs : Array
            Model outputs.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated mode for each output.
        """
        return super().mode(outputs, calibrated, **kwargs)

    def variance(
        self, outputs: jnp.ndarray, calibrated: bool = True, **kwargs
    ) -> jnp.ndarray:
        """
        Estimate the variance of the one-hot encoded target variable given the output, with respect to the predictive
        distribution.

        Parameters
        ----------
        outputs : jnp.ndarray
            Model outputs.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated variance for each output.
        """
        return super().variance(outputs, calibrated, **kwargs)

    def std(
        self,
        outputs: jnp.ndarray,
        variances: Optional[jnp.ndarray] = None,
        calibrated: bool = True,
    ) -> jnp.ndarray:
        """
        Estimate the standard deviation of the one-hot encoded target variable given the output, with respect to the
        predictive distribution.

        Parameters
        ----------
        outputs : jnp.ndarray
            Model outputs.
        variances: Optional[jnp.ndarray]
            Variance for each output.
        calibrated : bool
            Whether the outputs should be calibrated when computing this method. If `calibrated` is set to True, the
            model must have been calibrated beforehand.

        Returns
        -------
        jnp.ndarray
            The estimated standard deviation for each output.
        """
        return super().std(outputs, variances, calibrated)



from typing import Union

from jax import (
    random,
    vmap,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
from jax.scipy.stats import (
    multivariate_normal,
    norm,
)

from fortuna.distribution.base import Distribution
from fortuna.typing import Array


class DiagGaussian(Distribution):
    def __init__(self, mean: Union[float, Array], std: Union[float, Array]):
        """
        Diagonal Multivariate Gaussian class.

        :param mean: Union[float, Array]
            Mean parameter.
        :param std: Union[float, Array]
            Standard deviation parameter. If multi-dimensional, this represents the square-root of the diagonal of the
            covariance matrix.
        """
        self.mean = mean
        self.std = std
        self.dim = 1 if type(mean) in [int, float] else len(mean)

    def sample(self, rng: PRNGKeyArray, n_samples: int = 1) -> jnp.ndarray:
        """
        Sample from the diagonal Gaussian.

        :param rng: PRNGKeyArray
            Random number generator.
        :param n_samples: int
            Number of samples.

        :return: Array
            Samples. `shape = (number of samples, dimension)`
        """
        _, key = random.split(rng)
        return self.mean + self.std * random.normal(key, (n_samples, self.dim))

    def log_joint_prob(self, x: Union[float, Array]) -> Union[float, jnp.ndarray]:
        """
        Evaluate log-probability density function.

        :param x: Array
            Location(s) where the evaluation take place. Multiple locations allowed along the 0th axis.

        :return: Union[float, Array]
            Evaluation(s).
        """
        return jnp.sum(norm.logpdf(x, loc=self.mean, scale=self.std), -1)


class MultGaussian(Distribution):
    def __init__(self, mean: Array, cov: Array):
        """
        Multivariate Gaussian class.

        :param mean: Array
            Mean parameter.
        :param cov: Array
            Covariance matrix.
        """
        self.mean = mean
        self.cov = cov
        self.dim = len(mean)

    def sample(self, rng: PRNGKeyArray, n_samples: int = 1) -> jnp.ndarray:
        """
        Sample from the multivariate Gaussian.

        :param rng: PRNGKeyArray
            Random number generator.
        :param n_samples: int
            Number of samples.

        :return: Array
            Samples. `shape = (number of samples, dimension)`
        """
        keys = random.split(rng, n_samples)
        return vmap(lambda key: random.multivariate_normal(key, self.mean, self.cov))(
            keys
        )

    def log_joint_prob(self, x: Array) -> Union[float, jnp.ndarray]:
        """
        Evaluate log-probability density function.

        :param x: Array
            Location(s) where the evaluation take place. Multiple locations allowed along the 0th axis.

        :return: Union[float, Array]
            Evaluation(s).
        """
        return multivariate_normal.logpdf(x, mean=self.mean, cov=self.cov)



from copy import deepcopy
import os
from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from fortuna.training.mixin import WithCheckpointingMixin
from fortuna.training.train_state import TrainState
from fortuna.typing import (
    OptaxOptimizer,
    Path,
)


class TrainStateRepository(WithCheckpointingMixin):
    def __init__(self, checkpoint_dir: Optional[Path] = None):
        super().__init__()
        self.checkpoint_dir = checkpoint_dir
        self.__state = None

    def get(
        self,
        checkpoint_path: Optional[Path] = None,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> Union[Dict, TrainState]:
        if not checkpoint_path and not self.checkpoint_dir and not self.__state:
            raise ValueError("No state available.")
        if checkpoint_path or self.checkpoint_dir:
            return self.restore_checkpoint(
                restore_checkpoint_path=checkpoint_path or self.checkpoint_dir,
                optimizer=optimizer,
                prefix=prefix,
                **kwargs,
            )
        if optimizer is not None:
            self.__state = self.__state.replace(
                tx=optimizer, opt_state=optimizer.init(self.__state.params)
            )
        return deepcopy(self.__state)

    def put(
        self,
        state: TrainState,
        checkpoint_path: Optional[Path] = None,
        keep: int = 1,
        prefix: str = "checkpoint_",
    ) -> None:
        if checkpoint_path or self.checkpoint_dir:
            self.save_checkpoint(
                state=state,
                save_checkpoint_dir=checkpoint_path or self.checkpoint_dir,
                keep=keep,
                force_save=True,
                prefix=prefix,
            )
        else:
            self.__state = state

    def pull(
        self,
        checkpoint_path: Path = None,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> TrainState:
        state = self.get(
            checkpoint_path=checkpoint_path,
            optimizer=optimizer,
            prefix=prefix,
            **kwargs,
        )
        if checkpoint_path or self.checkpoint_dir:
            os.remove(
                checkpoint_path
                or self.get_path_latest_checkpoint(self.checkpoint_dir, prefix=prefix)
            )
        return state

    def update(
        self,
        variables: Dict,
        checkpoint_path: Path = None,
        optimizer: Optional[OptaxOptimizer] = None,
        keep: int = 1,
        prefix: str = "checkpoint_",
        **kwargs,
    ):
        state = self.pull(
            checkpoint_path=checkpoint_path,
            optimizer=optimizer,
            prefix=prefix,
            **kwargs,
        )
        state = state.replace(**variables)
        self.put(state, checkpoint_path=checkpoint_path, keep=keep, prefix=prefix)

    def extract(
        self,
        keys: List[str],
        checkpoint_path: Optional[Path] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> Dict:
        state = self.get(checkpoint_path=checkpoint_path, prefix=prefix, **kwargs)
        return {k: getattr(state, k) for k in keys}



from __future__ import annotations

from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
import jax.numpy as jnp

from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import (
    AnyKey,
    Array,
    Params,
)
from fortuna.utils.nested_dicts import nested_pair
from fortuna.utils.strings import (
    convert_string_to_jnp_array,
    encode_tuple_of_lists_of_strings_to_numpy,
)


class LaplaceState(PosteriorState):
    """
    Attributes
    ----------
    prior_log_var: float
        Prior log-variance value.
    encoded_name: jnp.ndarray
        Laplace state name encoded as an array.
    """

    prior_log_var: float = 0.0
    encoded_name: jnp.ndarray = convert_string_to_jnp_array("LaplaceState")
    _encoded_which_params: Optional[Dict[str, Array]] = None

    @classmethod
    def convert_from_map_state(
        cls,
        map_state: MAPState,
        hess_lik_diag: Union[Params, Tuple[Params, ...]],
        prior_log_var: Optional[float],
        which_params: Tuple[List[AnyKey], ...],
    ) -> LaplaceState:
        """
        Convert a MAP state into a Laplace state.

        Parameters
        ----------
        map_state: MAPState
            A MAP state.
        hess_lik_diag: Union[Params, Tuple[Params, ...]]
            Diagonal of the approximated Hessian of the likelihood.
        prior_log_var: float
            Prior log-variance value. If None, initialize it to 100.
        which_params: Tuple[List[AnyKey], ...]
            Sequences of keys pointing to the parameters over which `std` is defined. If `which_params` is None,
            `std` must be defined for all parameters.

        Returns
        -------
        LaplaceState
            A Laplace state instance.
        """
        params = map_state.params.unfreeze()
        if which_params is not None:
            params = nested_pair(
                d=params,
                key_paths=which_params,
                objs=hess_lik_diag,
                labels=("mean", "hess_lik_diag"),
            )
        else:
            for k, v in params.items():
                params[k] = FrozenDict(
                    {
                        "params": dict(
                            mean=v["params"], hess_lik_diag=hess_lik_diag[k]["params"]
                        )
                    }
                )
        d = vars(
            map_state.replace(
                params=FrozenDict(params), encoded_name=LaplaceState.encoded_name
            )
        )
        d["_encoded_which_params"] = encode_tuple_of_lists_of_strings_to_numpy(
            which_params
        )
        d["prior_log_var"] = prior_log_var if prior_log_var is not None else 0.0
        return LaplaceState.init_from_dict(d)



"""
Flax implementation of ResNet V1.
Taken as is from https://github.com/google/flax/blob/main/examples/imagenet/models.py
"""

# See issue #620.
# pytype: disable=wrong-arg-count

from functools import partial
from typing import (
    Any,
    Callable,
    Sequence,
    Tuple,
)

import flax.linen as nn
import jax.numpy as jnp

from fortuna.model.utils.spectral_norm import WithSpectralConv2DNorm
from fortuna.typing import Array

ModuleDef = Any


class ResNetBlock(nn.Module):
    """
    Residual network block.

    Attributes
    ----------
    filters: int
        Number of filters.
    conv: ModuleDef
        Convolution module.
    norm: ModuleDef
        Normalization module.
    activation: Callable
        Activation function.
    strides: Tuple[int, int]
        Strides.
    """

    filters: int
    conv: ModuleDef
    norm: ModuleDef
    activation: Callable
    strides: Tuple[int, int] = (1, 1)

    @nn.compact
    def __call__(
        self,
        x: jnp.ndarray,
    ) -> jnp.ndarray:
        """
        Block forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Block inputs.

        Returns
        -------
        jnp.ndarray
            Block outputs.
        """
        residual = x
        y = self.conv(self.filters, (3, 3), self.strides)(x)
        y = self.norm()(y)
        y = self.activation(y)
        y = self.conv(self.filters, (3, 3))(y)
        y = self.norm(scale_init=nn.initializers.zeros)(y)

        if residual.shape != y.shape:
            residual = self.conv(self.filters, (1, 1), self.strides, name="conv_proj")(
                residual
            )
            residual = self.norm(name="norm_proj")(residual)

        return self.activation(residual + y)


class BottleneckResNetBlock(nn.Module):
    """
    Bottleneck residual network block.

    Attributes
    ----------
    filters: int
        Number of filters.
    conv: ModuleDef
        Convolution module.
    norm: ModuleDef
        Normalization module.
    activation: Callable
        Activation function.
    strides: Tuple[int, int]
        Strides.
    """

    filters: int
    conv: ModuleDef
    norm: ModuleDef
    activation: Callable
    strides: Tuple[int, int] = (1, 1)

    @nn.compact
    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        """
        Bottleneck block forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Block inputs.

        Returns
        -------
        jnp.ndarray
            Block outputs.
        """
        residual = x
        y = self.conv(self.filters, (1, 1))(x)
        y = self.norm()(y)
        y = self.activation(y)
        y = self.conv(self.filters, (3, 3), self.strides)(y)
        y = self.norm()(y)
        y = self.activation(y)
        y = self.conv(self.filters * 4, (1, 1))(y)
        y = self.norm(scale_init=nn.initializers.zeros)(y)

        if residual.shape != y.shape:
            residual = self.conv(
                self.filters * 4, (1, 1), self.strides, name="conv_proj"
            )(residual)
            residual = self.norm(name="norm_proj")(residual)

        return self.activation(residual + y)


class DeepFeatureExtractorSubNet(nn.Module):
    """
     Deep feature extractor subnetwork.

    Attributes
    ----------
    stage_sizes: Sequence[int]
        Sizes for each stage.
    block_cls: ModuleDef
        Block class.
    num_filters: int
        Number of filters.
    dtype: Any
        Layers' dtype.
    activation: Callable
        Activation function.
    conv: ModuleDef
        Convolution module.
    """

    stage_sizes: Sequence[int]
    block_cls: ModuleDef
    num_filters: int = 64
    dtype: Any = jnp.float32
    activation: Callable = nn.relu
    conv: ModuleDef = nn.Conv

    @nn.compact
    def __call__(self, x: Array, train: bool = True) -> jnp.ndarray:
        """
        Deep feature extractor subnetwork forward pass.

        Parameters
        ----------
        x: Array
            Input data.
        train: bool
            Whether the call is performed during training.

        Returns
        -------
        jnp.ndarray
            Deep feature extractor representation.
        """
        if hasattr(self, "spectral_norm"):
            conv = self.spectral_norm(self.conv, train=train)
        else:
            conv = self.conv
        conv = partial(conv, use_bias=False, dtype=self.dtype)
        norm = partial(
            nn.BatchNorm,
            use_running_average=not train,
            momentum=0.9,
            epsilon=1e-5,
            dtype=self.dtype,
        )
        x = conv(
            self.num_filters, (7, 7), (2, 2), padding=[(3, 3), (3, 3)], name="conv_init"
        )(x)
        x = norm(name="bn_init")(x)
        x = nn.relu(x)
        x = nn.max_pool(x, (3, 3), strides=(2, 2), padding="SAME")
        for i, block_size in enumerate(self.stage_sizes):
            for j in range(block_size):
                strides = (2, 2) if i > 0 and j == 0 else (1, 1)
                x = self.block_cls(
                    self.num_filters * 2**i,
                    strides=strides,
                    conv=conv,
                    norm=norm,
                    activation=self.activation,
                )(x)
        x = jnp.mean(x, axis=(1, 2))
        return x


class OutputSubNet(nn.Module):
    """
    Output subnetwork.

    Attributes
    ----------
    output_dim: int
        Output dimension.
    """

    output_dim: int
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(self, x: jnp.ndarray, train: bool = True) -> jnp.ndarray:
        """
        Output subnetwork forward pass.

        Parameters
        ----------
        x: jnp.ndarray
            Deep feature extractor representation.
        train: bool
            Whether the call is performed during training.

        Returns
        -------
        jnp.ndarray
            Output of the subnetwork.
        """
        x = nn.Dense(self.output_dim, dtype=self.dtype)(x)
        x = jnp.asarray(x, self.dtype)
        return x


class ResNet(nn.Module):
    """
     Deep feature extractor subnetwork.

    Attributes
    ----------
    stage_sizes: Sequence[int]
        Sizes for each stage.
    block_cls: ModuleDef
        Block class.
    output_dim: int
        Output dimension.
    num_filters: int
        Number of filters.
    dtype: Any
        Layers' dtype.
    activation: Callable
        Activation function.
    conv: ModuleDef
        Convolution module.
    """

    stage_sizes: Sequence[int]
    block_cls: ModuleDef
    output_dim: int
    num_filters: int = 64
    dtype: Any = jnp.float32
    activation: Callable = nn.relu
    conv: ModuleDef = nn.Conv

    def setup(self):
        self.dfe_subnet = DeepFeatureExtractorSubNet(
            stage_sizes=self.stage_sizes,
            block_cls=self.block_cls,
            num_filters=self.num_filters,
            dtype=self.dtype,
            activation=self.activation,
            conv=self.conv,
        )
        self.output_subnet = OutputSubNet(output_dim=self.output_dim, dtype=self.dtype)

    def __call__(self, x: Array, train: bool = True) -> jnp.ndarray:
        """
        Forward pass.

        Parameters
        ----------
        x: Array
            Input data.
        train: bool
            Whether the call is performed during training.

        Returns
        -------
        jnp.ndarray
            Outputs.
        """
        x = self.dfe_subnet(x, train)
        x = self.output_subnet(x, train)
        return x


ResNet18 = partial(ResNet, stage_sizes=[2, 2, 2, 2], block_cls=ResNetBlock)
ResNet34 = partial(ResNet, stage_sizes=[3, 4, 6, 3], block_cls=ResNetBlock)
ResNet50 = partial(ResNet, stage_sizes=[3, 4, 6, 3], block_cls=BottleneckResNetBlock)
ResNet101 = partial(ResNet, stage_sizes=[3, 4, 23, 3], block_cls=BottleneckResNetBlock)
ResNet152 = partial(ResNet, stage_sizes=[3, 8, 36, 3], block_cls=BottleneckResNetBlock)
ResNet200 = partial(ResNet, stage_sizes=[3, 24, 36, 3], block_cls=BottleneckResNetBlock)


class ResNetDeepFeatureExtractorSubNetWithSN(
    WithSpectralConv2DNorm, DeepFeatureExtractorSubNet
):
    pass


# define the feature extractors with spectral norm
ResNet18DeepFeatureExtractorSubNetWithSN = partial(
    ResNetDeepFeatureExtractorSubNetWithSN,
    stage_sizes=[2, 2, 2, 2],
    block_cls=ResNetBlock,
)
ResNet34DeepFeatureExtractorSubNetWithSN = partial(
    ResNetDeepFeatureExtractorSubNetWithSN,
    stage_sizes=[3, 4, 6, 3],
    block_cls=ResNetBlock,
)
ResNet50DeepFeatureExtractorSubNetWithSN = partial(
    ResNetDeepFeatureExtractorSubNetWithSN,
    stage_sizes=[3, 4, 6, 3],
    block_cls=BottleneckResNetBlock,
)
ResNet101DeepFeatureExtractorSubNetWithSN = partial(
    ResNetDeepFeatureExtractorSubNetWithSN,
    stage_sizes=[3, 4, 23, 3],
    block_cls=BottleneckResNetBlock,
)
ResNet152DeepFeatureExtractorSubNetWithSN = partial(
    ResNetDeepFeatureExtractorSubNetWithSN,
    stage_sizes=[3, 8, 36, 3],
    block_cls=BottleneckResNetBlock,
)
ResNet200DeepFeatureExtractorSubNetWithSN = partial(
    ResNetDeepFeatureExtractorSubNetWithSN,
    stage_sizes=[3, 24, 36, 3],
    block_cls=BottleneckResNetBlock,
)



from typing import (
    Callable,
    Optional,
    Tuple,
)

import optax

from fortuna.typing import (
    AnyKey,
    Array,
    OptaxOptimizer,
)


class FitOptimizer:
    def __init__(
        self,
        method: Optional[OptaxOptimizer] = optax.adam(1e-3),
        n_epochs: int = 100,
        freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]] = None,
    ):
        """
        An object to configure the optimization in the posterior fitting.

        Parameters
        ----------
        method: OptaxOptimizer
            An Optax optimizer.
        n_epochs: int
            Maximum number of epochs to run the training for.
        freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]]
            A callable taking in input a path in the nested dictionary of parameters, as well as the corresponding
            array of parameters, and returns "trainable" or "freeze", according to whether the corresponding parameter
            should be optimized or not.
        """
        self.method = method
        self.n_epochs = n_epochs
        self.freeze_fun = freeze_fun



from typing import (
    List,
    Optional,
    Union,
)

from jax import (
    lax,
    random,
    vmap,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
import jax.scipy as jsp

from fortuna.data.loader import InputsLoader
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.predictive.base import Predictive
from fortuna.typing import Array


class RegressionPredictive(Predictive):
    def __init__(self, posterior: Posterior):
        """
        Regression predictive distribution class.

        Parameters
        ----------
        posterior : Posterior
             A posterior distribution object.
        """
        super().__init__(posterior)

    def mode(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        means: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        if means is not None:
            return means
        return self.mean(
            inputs_loader=inputs_loader,
            n_posterior_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )

    def aleatoric_entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        n_target_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive aleatoric entropy, that is

        .. math::
            -\mathbb{E}_{W|\mathcal{D}}[\mathbb{E}_{Y|W, x}[\log p(Y|W, x)]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_target_samples: int
            Number of target samples to draw for each input.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive aleatoric entropy for each input.
        """
        if rng is None:
            rng = self.rng.get()
        key1, *keys = random.split(rng, 1 + n_posterior_samples)

        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=key1,
            distribute=distribute,
        )

        ensemble_target_samples = lax.map(
            lambda variables: self.likelihood.prob_output_layer.sample(
                n_target_samples, variables[0], rng=variables[1]
            ),
            (ensemble_outputs, jnp.array(keys)),
        )

        def fun(i, _curr_sum):
            log_liks = self.likelihood.prob_output_layer.log_prob(
                ensemble_outputs[i], ensemble_target_samples[i]
            )
            _curr_sum -= jnp.mean(log_liks, 0)
            return _curr_sum

        curr_sum = fun(0, 0.0)
        curr_sum = lax.fori_loop(1, n_posterior_samples, fun, curr_sum)
        return curr_sum / n_posterior_samples

    def epistemic_entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        n_target_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive epistemic entropy, that is

        .. math::
            -\mathbb{E}_{Y|x, \mathcal{D}}[\log p(Y|x, \mathcal{D})] +
            \mathbb{E}_{W|\mathcal{D}}[\mathbb{E}_{Y|W, x}[\log p(Y|W, x)]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

         Note that the epistemic entropy above is defined as the difference between the predictive entropy and the
         aleatoric predictive entropy.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        n_target_samples: int
            Number of target samples to draw for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive epistemic entropy for each input.
        """
        if rng is None:
            rng = self.rng.get()
        key1, *keys = random.split(rng, 1 + n_posterior_samples)

        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=key1,
            distribute=distribute,
        )

        ensemble_target_samples = lax.map(
            lambda variables: self.likelihood.prob_output_layer.sample(
                n_target_samples, variables[0], rng=variables[1]
            ),
            (ensemble_outputs, jnp.array(keys)),
        )

        def fun(i, _curr_sum):
            @vmap
            def _log_pred_fun(target_sample: jnp.ndarray):
                logps = self.likelihood.prob_output_layer.log_prob(
                    ensemble_outputs, target_sample
                )
                return jsp.special.logsumexp(logps, 0) - jnp.log(n_posterior_samples)

            log_preds = _log_pred_fun(ensemble_target_samples[i])
            log_liks = self.likelihood.prob_output_layer.log_prob(
                ensemble_outputs[i], ensemble_target_samples[i]
            )
            _curr_sum -= jnp.mean(log_preds - log_liks, 0)
            return _curr_sum

        curr_sum = fun(0, 0.0)
        curr_sum = lax.fori_loop(1, n_posterior_samples, fun, curr_sum)
        return curr_sum / n_posterior_samples

    def entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        n_target_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive entropy, that is

        .. math::
            -\mathbb{E}_{Y|x, \mathcal{D}}[\log p(Y|x, \mathcal{D})],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_target_samples: int
            Number of target samples to draw for each input.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive entropy for each input.
        """
        if rng is None:
            rng = self.rng.get()
        key1, *keys = random.split(rng, 1 + n_posterior_samples)

        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=key1,
            distribute=distribute,
        )

        ensemble_target_samples = lax.map(
            lambda variables: self.likelihood.prob_output_layer.sample(
                n_target_samples, variables[0], rng=variables[1]
            ),
            (ensemble_outputs, jnp.array(keys)),
        )

        def fun(i, _curr_sum):
            @vmap
            def _log_pred_fun(target_sample: jnp.ndarray):
                logps = self.likelihood.prob_output_layer.log_prob(
                    ensemble_outputs, target_sample
                )
                return jsp.special.logsumexp(logps, 0) - jnp.log(n_posterior_samples)

            log_preds = _log_pred_fun(ensemble_target_samples[i])
            _curr_sum -= jnp.mean(log_preds, 0)
            return _curr_sum

        curr_sum = fun(0, 0.0)
        curr_sum = lax.fori_loop(1, n_posterior_samples, fun, curr_sum)
        return curr_sum / n_posterior_samples

    def credible_interval(
        self,
        inputs_loader: InputsLoader,
        n_target_samples: int = 30,
        error: float = 0.05,
        interval_type: str = "two-tailed",
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate credible intervals for the target variable. This is supported only if the target variable is scalar.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_target_samples: int
            Number of target samples to draw for each input.
        error: float
            The interval error. This must be a number between 0 and 1, extremes included. For example,
            `error=0.05` corresponds to a 95% level of credibility.
        interval_type: str
            The interval type. We support "two-tailed" (default), "right-tailed" and "left-tailed".
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            A credibility interval for each of the inputs.
        """
        supported_types = ["two-tailed", "right-tailed", "left-tailed"]
        if interval_type not in supported_types:
            raise ValueError(
                "`type={}` not recognised. Please choose among the following supported types: {}.".format(
                    interval_type, supported_types
                )
            )
        q = (
            jnp.array([0.5 * error, 1 - 0.5 * error])
            if interval_type == "two-tailed"
            else error
            if interval_type == "left-tailed"
            else 1 - error
        )
        qq = self.quantile(
            q=q,
            inputs_loader=inputs_loader,
            n_target_samples=n_target_samples,
            rng=rng,
            distribute=distribute,
        )
        if qq.shape[-1] != 1:
            raise ValueError(
                """Credibility intervals are only supported for scalar target variables."""
            )
        if interval_type == "two-tailed":
            lq, uq = qq.squeeze(2)
            return jnp.array(list(zip(lq, uq)))
        else:
            return qq

    def quantile(
        self,
        q: Union[float, Array, List],
        inputs_loader: InputsLoader,
        n_target_samples: Optional[int] = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> Union[float, jnp.ndarray]:
        r"""
        Estimate the `q`-th quantiles of the predictive probability density function.

        Parameters
        ----------
        q : Union[float, Array, List]
            Quantile or sequence of quantiles to compute. Each of these must be between 0 and 1, extremes included.
        inputs_loader : InputsLoader
            A loader of input data points.
        n_target_samples : int
            Number of target samples to sample for each input data point.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            Quantile estimate for each quantile and each input. If multiple quantiles `q` are given, the result's
            first axis is over different quantiles.
        """
        if type(q) == list:
            q = jnp.array(q)
        samples = self.sample(
            inputs_loader=inputs_loader,
            n_target_samples=n_target_samples,
            rng=rng,
            distribute=distribute,
        )
        return jnp.quantile(samples, q, axis=0)



from fortuna.output_calib_model.classification import OutputCalibClassifier
from fortuna.output_calib_model.config.base import Config
from fortuna.output_calib_model.config.checkpointer import Checkpointer
from fortuna.output_calib_model.config.monitor import Monitor
from fortuna.output_calib_model.config.optimizer import Optimizer
from fortuna.output_calib_model.config.processor import Processor
from fortuna.output_calib_model.regression import OutputCalibRegressor



from typing import List

import jax.numpy as jnp

from fortuna.typing import Array


class ConformalClassifier:
    """
    A base conformal classifier class.
    """

    def is_in(self, values: Array, conformal_sets: List) -> Array:
        """
        Check whether the values lie within their respective conformal sets.

        Parameters
        ----------
        values: Array
            Values to check if they lie in the respective conformal sets.
        conformal_sets: Array
            A conformal set for each input data point.

        Returns
        -------
        Array
            An array of ones or zero, indicating whether the values lie within their respective conformal sets.
        """
        return jnp.array([v in s for v, s in zip(values.tolist(), conformal_sets)])



import os
from typing import Optional

from flax.training import checkpoints

from fortuna.output_calib_model.state import OutputCalibState
from fortuna.training.mixin import WithCheckpointingMixin
from fortuna.typing import (
    OptaxOptimizer,
    Path,
)


class WithOutputCalibCheckpointingMixin(WithCheckpointingMixin):
    def restore_checkpoint(
        self,
        restore_checkpoint_path: Path,
        optimizer: Optional[OptaxOptimizer] = None,
        prefix: str = "checkpoint_",
        **kwargs,
    ) -> OutputCalibState:
        if not os.path.isdir(restore_checkpoint_path) and not os.path.isfile(
            restore_checkpoint_path
        ):
            raise ValueError(
                f"`restore_checkpoint_path={restore_checkpoint_path}` was not found."
            )
        d = checkpoints.restore_checkpoint(
            ckpt_dir=str(restore_checkpoint_path),
            target=None,
            step=None,
            prefix=prefix,
            parallel=True,
        )
        if d is None:
            raise ValueError(
                f"No checkpoint was found in `restore_checkpoint_path={restore_checkpoint_path}`."
            )

        return OutputCalibState.init_from_dict(d, optimizer, **kwargs)



from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)

from flax.core import FrozenDict
from jax import device_get
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp
from jax.tree_util import tree_map

from fortuna.prob_model.posterior.map.map_trainer import MAPTrainer
from fortuna.prob_model.posterior.swag.swag_state import SWAGState
from fortuna.training.callback import Callback
from fortuna.training.trainer import (
    JittedMixin,
    MultiDeviceMixin,
)
from fortuna.typing import (
    Array,
    Batch,
    Path,
)
from fortuna.utils.strings import encode_tuple_of_lists_of_strings_to_numpy


class SWAGTrainer(MAPTrainer):
    def __init__(self, *, which_params: Optional[Tuple[List[str]]], **kwargs):
        super(SWAGTrainer, self).__init__(**kwargs)
        self._mean_rav_params = None
        self._mean_squared_rav_params = None
        self._deviation_rav_params = None
        self._encoded_which_params = encode_tuple_of_lists_of_strings_to_numpy(
            which_params
        )

    def _update_state_with_stats(self, state: SWAGState) -> SWAGState:
        var = self._mean_squared_rav_params - self._mean_rav_params**2
        var = jnp.maximum(var, 0.0)
        return state.update(
            dict(
                mean=self._mean_rav_params
                if not self.multi_device
                else self._mean_rav_params[None],
                std=jnp.sqrt(var) if not self.multi_device else jnp.sqrt(var)[None],
                dev=self._deviation_rav_params
                if not self.multi_device
                else self._deviation_rav_params[None],
                _encoded_which_params=self._encoded_which_params
                if not self.multi_device
                else tree_map(lambda v: v[None], self._encoded_which_params),
            )
        )

    def training_step_end(
        self,
        current_epoch: int,
        state: SWAGState,
        aux: Dict[str, Any],
        batch: Batch,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
        callbacks: Optional[List[Callback]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[SWAGState, Dict[str, jnp.ndarray]]:
        if "rank" not in kwargs:
            raise AttributeError(
                """`rank` must be available in `kwargs` during training."""
            )
        rav_params = ravel_pytree(
            tree_map(lambda x: x[0], state.params)
            if self.multi_device
            else state.params
        )[0]
        if self._mean_rav_params is None:
            self._mean_rav_params = rav_params
            self._mean_squared_rav_params = rav_params**2
            self._deviation_rav_params = jnp.zeros((len(rav_params), 1))
        else:
            self._mean_rav_params *= current_epoch
            self._mean_rav_params += rav_params
            self._mean_rav_params /= current_epoch + 1

            self._mean_squared_rav_params *= current_epoch
            self._mean_squared_rav_params += rav_params**2
            self._mean_squared_rav_params /= current_epoch + 1
            self._deviation_rav_params = jnp.concatenate(
                (
                    self._deviation_rav_params[:, -kwargs["rank"] + 1 :],
                    (rav_params - self._mean_rav_params)[:, None],
                ),
                axis=1,
            )
        return super().training_step_end(
            current_epoch, state, aux, batch, metrics, callbacks, kwargs
        )

    def save_checkpoint(
        self,
        state: SWAGState,
        save_checkpoint_dir: Path,
        keep: int = 1,
        force_save: bool = False,
        prefix: str = "checkpoint_",
    ) -> None:
        state = self._update_state_with_stats(state)
        super().save_checkpoint(state, save_checkpoint_dir, keep, force_save, prefix)

    def on_train_end(self, state: SWAGState) -> SWAGState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )

        if self.freeze_fun is not None:
            state = state.replace(
                params=self._get_all_params(state), frozen_params=None
            )
        return self._update_state_with_stats(state)


class SWAGJittedMixin(JittedMixin):
    def on_train_end(self, state: SWAGState) -> SWAGState:
        state = super().on_train_end(state)
        return self._update_state_with_stats(state)


class SWAGMultiDeviceMixin(MultiDeviceMixin):
    def on_train_end(self, state: SWAGState) -> SWAGState:
        state = self._update_state_with_stats(state)
        return super().on_train_end(state)


class JittedSWAGTrainer(SWAGJittedMixin, SWAGTrainer):
    pass


class MultiDeviceSWAGTrainer(SWAGMultiDeviceMixin, SWAGTrainer):
    def on_train_end(self, state: SWAGState) -> SWAGState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )
        state = device_get(tree_map(lambda x: x[0], state))
        return self._update_state_with_stats(state)



from typing import Optional

from jax import (
    jit,
    lax,
    vmap,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
import jax.scipy as jsp
import numpy as np

from fortuna.data.loader import (
    ConcatenatedLoader,
    DataLoader,
    InputsLoader,
)
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.predictive.base import Predictive


class ClassificationPredictive(Predictive):
    def __init__(self, posterior: Posterior):
        """
        Classification predictive distribution class.

        Parameters
        ----------
        posterior : Posterior
             A posterior distribution object.
        """
        super().__init__(posterior)

    def mean(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive mean of the one-hot encoded target variable, that is

        .. math::
            \mathbb{E}_{\tilde{Y}|x, \mathcal{D}}[\tilde{Y}],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive mean for each input.
        """
        return super().mean(inputs_loader, n_posterior_samples, rng, distribute)

    def mode(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        means: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        if means is None:
            means = self.mean(
                inputs_loader=inputs_loader,
                n_posterior_samples=n_posterior_samples,
                rng=rng,
                distribute=distribute,
            )
        return jnp.argmax(means, -1)

    def aleatoric_variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive aleatoric variance of the one-hot encoded target variable, that is

        .. math::
            \text{Var}_{W|\mathcal{D}}[\mathbb{E}_{\tilde{Y}|W, x}[\tilde{Y}]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive aleatoric variance for each input.
        """
        return super().aleatoric_variance(
            inputs_loader, n_posterior_samples, rng, distribute
        )

    def epistemic_variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        **kwargs,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive epistemic variance of the one-hot encoded target variable, that is

        .. math::
            \mathbb{E}_{W|D}[\text{Var}_{\tilde{Y}|W, x}[\tilde{Y}]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive epistemic variance for each input.
        """
        return super().epistemic_variance(
            inputs_loader, n_posterior_samples, rng, distribute
        )

    def variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        aleatoric_variances: Optional[jnp.ndarray] = None,
        epistemic_variances: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive variance of the one-hot encoded target variable, that is

        .. math::
            \text{Var}_{\tilde{Y}|x, D}[\tilde{Y}],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        aleatoric_variances: Optional[jnp.ndarray]
            An estimate of the aleatoric predictive variance.
        epistemic_variances: Optional[jnp.ndarray]
            An estimate of the epistemic predictive variance.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive variance for each input.
        """
        return super().variance(
            inputs_loader,
            n_posterior_samples,
            aleatoric_variances,
            epistemic_variances,
            rng,
            distribute,
        )

    def std(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        variances: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive standard deviation of the one-hot encoded target variable, that is

        .. math::
            \sqrt{\text{Var}_{\tilde{Y}|x, D}[\tilde{Y}]},

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        variances: Optional[jnp.ndarray]
            An estimate of the predictive variance.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive standard deviation for each input.
        """
        return super().std(
            inputs_loader, n_posterior_samples, variances, rng, distribute
        )

    def aleatoric_entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive aleatoric entropy, that is

        .. math::
            -\mathbb{E}_{W|\mathcal{D}}[\mathbb{E}_{Y|W, x}[\log p(Y|W, x)]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive aleatoric entropy for each input.
        """
        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )
        n_classes = ensemble_outputs.shape[-1]

        @jit
        def _entropy_term(i: int):
            targets = i * jnp.ones(ensemble_outputs.shape[1])

            def _log_lik_fun(outputs):
                return self.likelihood.prob_output_layer.log_prob(outputs, targets)

            log_liks = vmap(_log_lik_fun)(ensemble_outputs)
            return jnp.mean(jnp.exp(log_liks) * log_liks, 0)

        return -jnp.sum(vmap(_entropy_term)(jnp.arange(n_classes)), 0)

    def epistemic_entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive epistemic entropy, that is

        .. math::
            -\mathbb{E}_{Y|x, \mathcal{D}}[\log p(Y|x, \mathcal{D})] +
            \mathbb{E}_{W|\mathcal{D}}[\mathbb{E}_{Y|W, x}[\log p(Y|W, x)]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

         Note that the epistemic entropy above is defined as the difference between the predictive entropy and the
         aleatoric predictive entropy.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive epistemic entropy for each input.
        """
        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )
        n_classes = ensemble_outputs.shape[-1]

        @jit
        def _entropy_term(i: int):
            targets = i * jnp.ones(ensemble_outputs.shape[1])

            def _log_lik_fun(outputs):
                return self.likelihood.prob_output_layer.log_prob(outputs, targets)

            log_liks = vmap(_log_lik_fun)(ensemble_outputs)
            log_preds = jsp.special.logsumexp(log_liks, 0) - jnp.log(
                n_posterior_samples
            )
            return jnp.exp(log_preds) * log_preds - jnp.mean(
                jnp.exp(log_liks) * log_liks, 0
            )

        return -jnp.sum(vmap(_entropy_term)(jnp.arange(n_classes)), 0)

    def entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive entropy, that is

        .. math::
            -\mathbb{E}_{Y|x, \mathcal{D}}[\log p(Y|x, \mathcal{D})],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive entropy for each input.
        """
        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )
        n_classes = ensemble_outputs.shape[-1]

        @jit
        def _entropy_term(i: int):
            targets = i * jnp.ones(ensemble_outputs.shape[1])

            def _log_lik_fun(outputs):
                return self.likelihood.prob_output_layer.log_prob(outputs, targets)

            log_liks = vmap(_log_lik_fun)(ensemble_outputs)
            log_preds = jsp.special.logsumexp(log_liks, 0) - jnp.log(
                n_posterior_samples
            )
            return jnp.exp(log_preds) * log_preds

        return -jnp.sum(vmap(_entropy_term)(jnp.arange(n_classes)), 0)

    def credible_set(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        error: float = 0.05,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate credible sets for the target variable. This is done by sorting the class probabilities in descending order
        and including classes until the sum > 1-error.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples: int
            Number of posterior samples to draw for each input.
        error: float
            The set error. This must be a number between 0 and 1, extremes included. For example,
            `error=0.05` corresponds to a 95% level of credibility.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            A credibility set for each of the inputs.
        """
        p_classes = self.mean(
            inputs_loader=inputs_loader,
            n_posterior_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )
        n_classes = jnp.shape(p_classes)[1]
        labels = jnp.argsort(p_classes, axis=-1)[:, ::-1]
        p_classes_sorted = jnp.sort(p_classes, axis=-1)[:, ::-1]
        region_classes = jnp.cumsum(p_classes_sorted, axis=1) > (1 - error)

        # Convert CB region into sets
        index_true = jnp.argmax(region_classes, axis=1)  # first index where True
        credible_set = np.zeros(len(index_true), dtype=object)
        for s in np.arange(n_classes):
            idx = np.where(index_true == s)[0]
            credible_set[idx] = labels[idx, : s + 1].tolist()

        return credible_set

    def conformal_set(
        self,
        train_data_loader: DataLoader,
        test_inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        error: float = 0.05,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        return_ess: bool = False,
    ) -> jnp.ndarray:
        r"""
        Estimate conformal sets for the target variable.

        Parameters
        ----------
        train_data_loader : DataLoader
            A training data loader.
        test_inputs_loader : InputsLoader
            A test inputs loader.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        error: float
            The set error. This must be a number between 0 and 1, extremes included. For example,
            `error=0.05` corresponds to a 95% level of confidence.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.
        return_ess: bool
            Whether to compute effective sample size of importance weights or not.

        Returns
        -------
        List[List[int]]
            A list of conformal sets for each test input.
        """

        # Extract training data
        n_train = train_data_loader.size
        n_classes = train_data_loader.num_unique_labels

        # Extract test inputs and evaluate on each class (e.g. Y = 0, Y = 1)
        n_test = test_inputs_loader.size
        test_data_grid_loader = DataLoader.from_inputs_loaders(
            inputs_loaders=[test_inputs_loader] * n_classes,
            targets=jnp.arange(n_classes).tolist(),
            how="concatenate",
        )

        # Combine training data and test data grid into single loader, so random posterior samples are the same rng
        train_test_data_loader = ConcatenatedLoader(
            loaders=[train_data_loader, test_data_grid_loader]
        )

        # returns n_posterior_samples x (n_classes*n_test +n)
        ensemble_train_test_log_probs = self.ensemble_log_prob(
            data_loader=train_test_data_loader,
            n_posterior_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )
        # Split training log_probs
        ensemble_train_log_probs = ensemble_train_test_log_probs[
            :, 0:n_train
        ]  # training log likelihood

        # test log prob for each test input, posterior sample, and class (shape =  n_test x  n_posterior_samples x n_classes)
        ensemble_testgrid_log_probs = jnp.dstack(
            jnp.vsplit(ensemble_train_test_log_probs[:, n_train:].T, n_classes)
        )  # Split test log_probs for each class (in third axis)

        @jit  # compute rank of nonconformity score (unnormalized by n+1)
        def _compute_cb_region_importancesampling(
            ensemble_testgrid_log_probs,
        ):
            # compute importance sampling weights and normalizing constant
            importance_weights = jnp.exp(ensemble_testgrid_log_probs.T)
            normalizing_constant = jnp.sum(importance_weights, axis=1).reshape(-1, 1)

            # compute predictives for y_i,x_i and y_new,x_n+1
            prob_train = jnp.dot(
                importance_weights / normalizing_constant,
                jnp.exp(ensemble_train_log_probs),
            )
            prob_test = (
                jnp.sum(importance_weights**2, axis=1).reshape(-1, 1)
                / normalizing_constant
            )

            # compute nonconformity score and sort
            prob_train_test = jnp.concatenate((prob_train, prob_test), axis=1)
            rank_test = jnp.sum(
                prob_train_test <= prob_train_test[:, -1].reshape(-1, 1), axis=1
            )

            # Compute region of grid which is in confidence set
            region_true = rank_test > error * (n_train + 1)

            return region_true

        # Compute CB region for each test input
        conformal_region = np.array(
            lax.map(_compute_cb_region_importancesampling, ensemble_testgrid_log_probs)
        )

        # Convert CB region into sets
        sizes = (conformal_region.sum(axis=1)).astype("int32")
        region_argsort = np.argsort(conformal_region, axis=1)[:, ::-1]
        conformal_set = np.zeros(len(sizes), dtype=object)
        for s in np.unique(sizes):
            idx = np.where(sizes == s)[0]
            conformal_set[idx] = region_argsort[idx, :s][:, ::-1].tolist()

        if return_ess:
            ## DIAGNOSE IMPORTANCE WEIGHTS ##
            @jit  # compute Effective sample size
            def _diagnose_importancesampling_weights(
                ensemble_testgrid_log_probs,
            ):
                # compute importance sampling weights and normalizing
                log_importance_weights = ensemble_testgrid_log_probs.T.reshape(
                    jnp.shape(ensemble_testgrid_log_probs)[1], -1, 1
                )
                log_normalizing_constant = jsp.special.logsumexp(
                    log_importance_weights, axis=1
                )

                # compute ESS
                importance_weights = jnp.exp(
                    log_importance_weights - log_normalizing_constant.reshape(-1, 1, 1)
                )
                ESS = 1 / jnp.sum(importance_weights**2, axis=1)
                return ESS

            ###
            ESS = lax.map(
                _diagnose_importancesampling_weights, ensemble_testgrid_log_probs
            )
            return conformal_set, ESS

        else:
            return conformal_set



# implementation adapted from https://github.com/google/edward2/blob/main/edward2/jax/nn/random_feature.py

import dataclasses
import functools
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Tuple,
    Type,
)

import flax.linen as nn
from jax import (
    lax,
    random,
)
import jax.numpy as jnp
from jax.random import PRNGKeyArray

from fortuna.typing import (
    Array,
    Shape,
)

linalg = lax.linalg

# Default config for random features.
default_rbf_bias_init = nn.initializers.uniform(scale=2.0 * jnp.pi)
# Using "he_normal" style random feature distribution (see https://arxiv.org/abs/1502.01852).
# Effectively, this is equivalent to approximating a RBF kernel but with the input standardized by
# its dimensionality (i.e., input_scaled = input * sqrt(2. / dim_input)) and
# empirically leads to better performance for neural network inputs.
# default_rbf_kernel_init = nn.initializers.variance_scaling(
#     scale=2.0, mode="fan_in", distribution="normal"
# )
default_rbf_kernel_init = nn.initializers.normal(stddev=1.0)

# Default field value for kwargs, to be used for data class declaration.
default_kwarg_dict = lambda: dataclasses.field(default_factory=dict)

SUPPORTED_LIKELIHOOD = ("binary_logistic", "poisson", "gaussian")


class RandomFeatureGaussianProcess(nn.Module):
    """
    A Gaussian process layer using random Fourier Features.

    See `Simple and Principled Uncertainty Estimation with Deterministic
    Deep Learning via Distance Awareness <https://arxiv.org/abs/2006.10108>`_

    Attributes
    ----------
    features: int
        The number of output units.
    hidden_features: int
        The number of hidden random fourier features.
    normalize_input: bool
        Whether to normalize the input using nn.LayerNorm.
    norm_kwargs: Mapping[str, Any]
        Optional keyword arguments to the input nn.LayerNorm layer.
    hidden_kwargs: Mapping[str, Any]
        Optional keyword arguments to the random feature layer.
    output_kwargs: Mapping[str, Any]
        Optional keyword arguments to the predictive logit layer.
    covariance_kwargs: Mapping[str, Any]
        Optional keyword arguments to the predictive covariance layer.
    """

    features: int
    hidden_features: int = 1024
    normalize_input: bool = False

    # Optional keyword arguments.
    norm_kwargs: Mapping[str, Any] = default_kwarg_dict()
    hidden_kwargs: Mapping[str, Any] = default_kwarg_dict()
    output_kwargs: Mapping[str, Any] = default_kwarg_dict()
    covariance_kwargs: Mapping[str, Any] = default_kwarg_dict()

    def setup(self):
        # pylint:disable=invalid-name,not-a-mapping
        if self.normalize_input:
            # Prefer a parameter-free version of LayerNorm by default
            # (see `Xu et al., 2019 <https://papers.nips.cc/paper/2019/file/2f4fe03d77724a7217006e5d16728874-Paper.pdf>`_)
            # Can be overwritten by passing norm_kwargs=dict(use_bias=..., use_scales=...).
            LayerNorm = functools.partial(nn.LayerNorm, use_bias=False, use_scale=False)
            self.sngp_norm_layer = LayerNorm(**self.norm_kwargs)

        self.sngp_random_features_layer = RandomFourierFeatures(
            features=self.hidden_features,
            **self.hidden_kwargs,
        )
        self.sngp_dense_layer = nn.Dense(features=self.features, **self.output_kwargs)
        self.sngp_covariance_layer = LaplaceRandomFeatureCovariance(
            hidden_features=self.hidden_features, **self.covariance_kwargs
        )
        # pylint:enable=invalid-name,not-a-mapping

    def __call__(
        self,
        inputs: Array,
        return_full_covariance: bool = False,
    ) -> Tuple[Array, Array]:
        """
        Computes Gaussian process outputs.

        Parameters
        ----------
        inputs: Array
            The nd-array of shape (batch_size, ..., input_dim).
        return_full_covariance: bool
            Whether to return the full covariance matrix, shape
            (batch_size, batch_size), or only return the predictive variances with
            shape (batch_size, ).

        Returns
        -------
        Tuple[Array, Array]
          A tuple of predictive logits, predictive covariance and (optionally)
          random Fourier features.
        """
        gp_inputs = self.sngp_norm_layer(inputs) if self.normalize_input else inputs
        gp_features = self.sngp_random_features_layer(gp_inputs)

        gp_logits = self.sngp_dense_layer(gp_features)
        gp_covariance = self.sngp_covariance_layer(
            gp_features, gp_logits, diagonal_only=not return_full_covariance
        )
        return gp_logits, gp_covariance


class RandomFourierFeatures(nn.Module):
    """
    A random fourier feature (RFF) layer that approximates a kernel model.

    The random feature transformation is a one-hidden-layer network with
    non-trainable weights (see, e.g., Algorithm 1 of
    `Random Features for Large-Scale
    Kernel Machines <https://people.eecs.berkeley.edu/~brecht/papers/07.rah.rec.nips.pdf>`_):

    .. math::
        f(x) = \gamma * cos(\mathbf{W}\mathbf{x} + \mathbf{b})

    where :math:`\mathbf{W}` is the kernel matrix, :math:`\mathbf{b}` is the bias
    and :math:`\gamma` is the output scale.
    The forward pass logic closely follows that of the `nn.Dense` layer.

    Attributes
    ----------
    features: int
        The number of output units.
    feature_scale: Optional[float]
        Scale to apply to the output.
        When using GP layer as the output layer of a nerual network, it is recommended to set this to 1.
        to prevent it from changing the learning rate to the hidden layers.
    kernel_init:  Callable[[PRNGKeyArray, Shape, Type], Array]
         Callable[[PRNGKeyArray, Shape, Type], Array] function for the weight matrix.
    bias_init:  Callable[[PRNGKeyArray, Shape, Type], Array]
         Callable[[PRNGKeyArray, Shape, Type], Array] function for the bias.
    seed: int
        Random seed for generating random features. This will override the external RNGs.
    dtype: Type
        The dtype of the computation.
    """

    features: int
    kernel_scale: Optional[float] = 1.0
    feature_scale: Optional[float] = 1.0
    kernel_init: Callable[[PRNGKeyArray, Shape, Type], Array] = default_rbf_kernel_init
    bias_init: Callable[[PRNGKeyArray, Shape, Type], Array] = default_rbf_bias_init
    seed: int = 0
    dtype: Type = jnp.float32
    collection_name: str = "random_features"

    def setup(self):
        # Defines the random number generator.
        self.rng = random.PRNGKey(self.seed)

        # Processes random feature scale.
        self._feature_scale = self.feature_scale
        if self._feature_scale is None:
            self._feature_scale = jnp.sqrt(2.0 / self.features)
        self._feature_scale = jnp.asarray(self._feature_scale, dtype=self.dtype)

    @nn.compact
    def __call__(self, inputs: Array) -> Array:
        """
        Applies random feature transformation along the last dimension of inputs.

        Parameters
        ----------
        inputs: Array
            The nd-array to be transformed.

        Returns
        -------
        Array
          The transformed input.
        """
        # Initializes variables.
        input_dim = inputs.shape[-1]

        kernel_rng, bias_rng = random.split(self.rng, num=2)
        kernel_shape = (input_dim, self.features)

        kernel = self.variable(
            self.collection_name,
            "kernel",
            self.kernel_init,
            kernel_rng,
            kernel_shape,
            self.dtype,
        )
        kernel_scale = self.param(
            "rf_kernel_scale",
            nn.initializers.constant(self.kernel_scale),
            (1,),
            self.dtype,
        )
        bias = self.variable(
            self.collection_name,
            "bias",
            self.bias_init,
            bias_rng,
            (self.features,),
            self.dtype,
        )

        # Specifies multiplication dimension.
        contracting_dims = ((inputs.ndim - 1,), (0,))
        batch_dims = ((), ())

        # Performs forward pass.
        inputs = jnp.asarray(inputs, self.dtype)

        outputs = lax.dot_general(
            inputs, (1.0 / kernel_scale) * kernel.value, (contracting_dims, batch_dims)
        )
        outputs = outputs + jnp.broadcast_to(bias.value, outputs.shape)

        return self._feature_scale * jnp.cos(outputs)


class LaplaceRandomFeatureCovariance(nn.Module):
    """
    Computes the approximated posterior covariance using Laplace method.

    Attributes
    ----------
    hidden_features: int
        The number of random fourier features.
    ridge_penalty: float
        Initial Ridge penalty to weight covariance matrix.
        This value is used to stabilize the eigenvalues of weight covariance estimate :math:`\Sigma` so that
        the matrix inverse can be computed for :math:`\Sigma = (\mathbf{I}*s+\mathbf{X}^T\mathbf{X})^{-1}`.
        The ridge factor :math:`s` cannot be too large since otherwise it will dominate
        making the covariance estimate not meaningful.
    momentum: Optional[float]
        A discount factor used to compute the moving average for posterior
        precision matrix. Analogous to the momentum factor in batch normalization.
        If `None` then update covariance matrix using a naive sum without
        momentum, which is desirable if the goal is to compute the exact
        covariance matrix by passing through data once (say in the final epoch).
        In this case, make sure to reset the precision matrix variable between
        epochs to avoid double counting.
    dtype: Type
        The dtype of the computation
    """

    hidden_features: int
    ridge_penalty: float = 1.0
    momentum: Optional[float] = None
    collection_name: str = "laplace_covariance"
    dtype: Type = jnp.float32

    def setup(self):
        if self.momentum is not None:
            if self.momentum < 0.0 or self.momentum > 1.0:
                raise ValueError(
                    f"`momentum` must be between (0, 1). " f"Got {self.momentum}."
                )

    @nn.compact
    def __call__(
        self,
        gp_features: Array,
        gp_logits: Optional[Array] = None,
        diagonal_only: bool = True,
    ) -> Optional[Array]:
        """
        Updates the precision matrix and computes the predictive covariance.

        NOTE:
        The precision matrix will be updated only during training (i.e., when
        `self.collection_name` are in the list of mutable variables). The covariance
        matrix will be computed only during inference to avoid repeated calls to the
        (expensive) `linalg.inv` op.

        Parameters
        ----------
        gp_features: Array
            The nd-array of random fourier features, shape (batch_size, ..., hidden_features).
        gp_logits: Optional[Array]
            The nd-array of predictive logits, shape (batch_size, ..., logit_dim).
            Cannot be None.
        diagonal_only: bool
            Whether to return only the diagonal elements of the predictive covariance matrix (i.e., the predictive variance).

        Returns
        -------
        Optional[Array]
          The predictive variances of shape (batch_size, ) if diagonal_only=True,
          otherwise the predictive covariance matrix of shape
          (batch_size, batch_size).
        """
        gp_features = jnp.asarray(gp_features, self.dtype)

        # Flatten GP features and logits to 2-d, by doing so we treat all the
        # non-final dimensions as the batch dimensions.
        gp_features = jnp.reshape(gp_features, [-1, self.hidden_features])

        if gp_logits is not None:
            gp_logits = jnp.asarray(gp_logits, self.dtype)
            gp_logits = jnp.reshape(gp_logits, [gp_features.shape[0], -1])

        precision_matrix = self.variable(
            self.collection_name,
            "precision_matrix",
            lambda: self.initial_precision_matrix(),
        )  # pylint: disable=unnecessary-lambda

        # Updates the precision matrix during training.
        initializing = self.is_mutable_collection("params")
        training = self.is_mutable_collection(self.collection_name)

        if training and not initializing:
            precision_matrix.value = self.update_precision_matrix(
                gp_features, gp_logits, precision_matrix.value
            )

        # Computes covariance matrix during inference.
        if not training:
            return self.compute_predictive_covariance(
                gp_features, precision_matrix, diagonal_only
            )

    def initial_precision_matrix(self):
        """Returns the initial diagonal precision matrix."""
        return jnp.eye(self.hidden_features, dtype=self.dtype) * self.ridge_penalty

    def update_precision_matrix(
        self, gp_features: Array, gp_logits: Optional[Array], precision_matrix: Array
    ) -> Array:
        """Updates precision matrix given a new batch.

        Parameters
        ----------
          gp_features: Array
            Random features from the new batch, shape (batch_size, hidden_features)
          gp_logits: Optional[Array]
            Predictive logits from the new batch, shape (batch_size, logit_dim).
            Currently only `logit_dim=1` is supported.
          precision_matrix: Array
            The current precision matrix, shape (hidden_features, hidden_features).

        Returns
        -------
        Array
            Updated precision matrix, shape (hidden_features, hidden_features).
        """
        # Computes precision matrix within new batch.
        prob_multiplier = 1.0

        gp_features_adj = jnp.sqrt(prob_multiplier) * gp_features
        batch_prec_mat = jnp.matmul(jnp.transpose(gp_features_adj), gp_features_adj)

        # Updates precision matrix.
        if self.momentum is None:
            # Performs exact update without momentum.
            precision_matrix_updated = precision_matrix + batch_prec_mat
        else:
            batch_size = gp_features.shape[0]
            precision_matrix_updated = (
                self.momentum * precision_matrix
                + (1 - self.momentum) * batch_prec_mat / batch_size
            )
        return precision_matrix_updated

    def compute_predictive_covariance(
        self, gp_features: Array, precision_matrix: nn.Variable, diagonal_only: bool
    ) -> Array:
        """
        Computes the predictive covariance.

        Approximates the Gaussian process posterior using random features.
        Given training random feature :math:`\mathbf{\Phi_{tr}}` (num_train, num_hidden) and testing
        random feature :math:`\mathbf{\Phi_{ts}}` (batch_size, num_hidden). The predictive covariance
        matrix is computed as (assuming Gaussian likelihood):
        :math:`s * \mathbf{\Phi_{ts}}(\mathbf{I}*s + \mathbf{\Phi_{tr}}^{T}*\mathbf{\Phi_{tr}})^{-1}\mathbf{\Phi_{tr}}^{^T}`

        where :math:`s` is the ridge factor to be used for stablizing the inverse, and \mathbf{I} is
        the identity matrix with shape (num_hidden, num_hidden). The above
        description is formal only: the actual implementation uses a Cholesky
        factorization of the covariance matrix.

        Parameters
        ----------
        gp_features: Array
            The random feature of testing data to be used for computing the covariance matrix.
            Shape (batch_size, gp_hidden_size).
        precision_matrix: nn.Variable
            The model's precision matrix.
        diagonal_only: bool
            Whether to return only the diagonal elements of the predictive covariance matrix
            (i.e., the predictive variances).

        Returns
        -------
        Array
            The predictive variances of shape (batch_size, ) if `diagonal_only=True`,
            otherwise the predictive covariance matrix of shape (batch_size, batch_size).
        """
        chol = linalg.cholesky(precision_matrix.value)
        chol_t_cov_feature_product = linalg.triangular_solve(
            chol, gp_features.T, left_side=True, lower=True
        )

        if diagonal_only:
            # Compute diagonal element only, shape (batch_size, ).
            gp_covar = jnp.square(chol_t_cov_feature_product).sum(0)
        else:
            # Compute full covariance matrix, shape (batch_size, batch_size).
            gp_covar = chol_t_cov_feature_product.T @ chol_t_cov_feature_product

        return self.ridge_penalty * gp_covar



from typing import (
    Optional,
    Union,
)

import jax.numpy as jnp

from fortuna.typing import Array


class MultivalidMethod:
    def __init__(self, seed: int = 0):
        """
        A base multivalid method.

        Parameters
        ----------
        seed: int
            Random seed.
        """
        self._seed = seed
        self._patches = None
        self._n_buckets = None

    @property
    def patches(self):
        return self._patches

    @property
    def n_buckets(self):
        return self._n_buckets

    @n_buckets.setter
    def n_buckets(self, n_buckets):
        self._n_buckets = n_buckets

    @staticmethod
    def _get_buckets(n_buckets: int):
        return jnp.linspace(0, 1, n_buckets)

    @staticmethod
    def _round_to_buckets(v: Array, buckets: Array):
        return buckets[jnp.argmin(jnp.abs(v - buckets))]

    @staticmethod
    def _maybe_check_values(
        values: Optional[Array], test_values: Optional[Array] = None
    ):
        if values is not None:
            if values.ndim != 1:
                raise ValueError("`values` must be a 1-dimensional array.")
            if values is not None and jnp.any(values < 0) or jnp.any(values > 1):
                raise ValueError("All elements in `values` must be within [0, 1].")
        if test_values is not None:
            if jnp.any(test_values < 0) or jnp.any(test_values > 1):
                raise ValueError("All elements in `test_values` must be within [0, 1].")

    @staticmethod
    def _check_scores(scores: Array):
        if scores.ndim != 1:
            raise ValueError("`scores` must be a 1-dimensional array.")
        if jnp.any(scores < 0) or jnp.any(scores > 1):
            raise ValueError("All elements in `scores` must be within [0, 1].")

    @staticmethod
    def _process_scores(scores: Array):
        scores = jnp.copy(scores)
        if scores.ndim == 1:
            scores = jnp.copy(scores[:, None])
        return scores

    @staticmethod
    def _maybe_init_min_prob_b(
        min_prob_b: Union[str, float], n_buckets: int, n_dims: int
    ):
        if min_prob_b == "auto":
            return 1 / (n_buckets * n_dims)
        return min_prob_b

    @staticmethod
    def _maybe_process_values(
        values: Optional[Array], test_values: Optional[Array]
    ) -> Optional[Array]:
        if values is not None:
            values = jnp.copy(values)
        if test_values is not None:
            test_values = jnp.copy(test_values)
        return values, test_values



from fortuna.calib_model.classification import CalibClassifier
from fortuna.calib_model.config.base import Config
from fortuna.calib_model.config.checkpointer import Checkpointer
from fortuna.calib_model.config.hyperparameters import Hyperparameters
from fortuna.calib_model.config.monitor import Monitor
from fortuna.calib_model.config.optimizer import Optimizer
from fortuna.calib_model.config.processor import Processor
from fortuna.calib_model.regression import CalibRegressor
